# NEW -- REFACTOR FROM main.py and get_factors to Models. 
import logging
from typing import Any
import pyarrow as pa
from getfactormodels import models as factor_models
from getfactormodels.models.registry import resolve_model_key
from getfactormodels.utils.arrow_utils import (
    print_table_preview,
    rearrange_columns,
)

log = logging.getLogger(__name__)

class Models:
    """Interface to instantiate and join multiple factor models."""
    def __init__(self, model_ids: str | list[str], **kwargs):
        self._input_ids = [model_ids] if isinstance(model_ids, (str, int)) else model_ids
        self.params = kwargs
        self._data: pa.Table | None = None

    def _get_model_class(self, key: str):
        """Maps registry keys to concrete Python classes."""
        mapping = {
            "3": "FamaFrenchFactors", "5": "FamaFrenchFactors", "6": "FamaFrenchFactors",
            "4": "CarhartFactors", "Q": "QFactors", "Qclassic": "QFactors",
            "HMLDevil": "HMLDevilFactors",
        }
        class_name = mapping.get(key, f"{key}Factors")
        
        if not hasattr(factor_models, class_name):
            raise ValueError(f"Class '{class_name}' not found in factor_models.")
        return getattr(factor_models, class_name)

    def load(self, client=None) -> "Models":
        """The coordination step: triggers worker loads and joins them."""
        tables = []
        
        for mid in self._input_ids:
            key = resolve_model_key(mid)
            cls = self._get_model_class(key)
            
            # Inject model-specific params into the kwargs
            inst_params = self.params.copy()
            if key in ("3", "5", "6"): inst_params["model"] = key
            if key == "Qclassic": inst_params["classic"] = True
            
            # This calls the Worker's load()
            instance = cls(**inst_params).load(client=client)
            tables.append(instance.data)

        if not tables:
            raise RuntimeError("No data could be loaded.")

        # Join Logic: combine multiple tables into one
        combined = tables[0]
        for next_table in tables[1:]:
            # Drop overlapping columns to avoid duplicate factor names (SMB, HML, etc)
            overlap = set(combined.column_names) & set(next_table.column_names)
            overlap.discard('date')
            if overlap:
                next_table = next_table.drop(list(overlap))
                
            combined = combined.join(next_table, keys='date', join_type='left outer')

        self._data = rearrange_columns(combined)
        return self

    @property
    def data(self) -> pa.Table:
        if self._data is None: 
            self.load()
        return self._data

    def to_pandas(self):
        return self.data.to_pandas().set_index('date')

    def __str__(self):
        if self._data is None: return "Models(Not Loaded)"
        header = f"Joined Factor Models: {', '.join(map(str, self._input_ids))}\n"
        return header + print_table_preview(self._data)



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import logging
import pyarrow as pa
import pyarrow.compute as pc

"""Utilities for working with pyarrow."""

log = logging.getLogger(__name__)


def scale_to_decimal(table: pa.Table) -> pa.Table:
    """Standardize float cols to decimal (5.2 -> 0.052)."""
    for i, field in enumerate(table.schema):
        if pa.types.is_floating(field.type):
            scaled_col = pc.divide(table.column(i), 100.0)
            table = table.set_column(i, field.name, scaled_col)
    return table


def round_to_precision(table: pa.Table, precision: int) -> pa.Table:
    """Rounds all float cols in a pa.Table to precision, and RF to 4.

    Note:
    Used by all models.
    """
    # auto resolve to: precision arg, else model's ._precision or '6'.
    prec_val = precision or getattr(table, '_precision', 6)
    rf_cols = {'RF', 'R_F'} # TODO: fix this up.

    new_cols = []
    for i, field in enumerate(table.schema):
        col = table.column(i)
        if pa.types.is_floating(field.type):
            p = 4 if field.name.upper() in rf_cols else 6 if field.name.upper() == "RF_AQR" else prec_val
            col = pc.round(col, ndigits=p, round_mode='half_to_even')
        new_cols.append(col)

    return pa.Table.from_arrays(new_cols, schema=table.schema)


def rearrange_columns(table: pa.Table) -> pa.Table:
    """Standardize column orders: 'date', [FACTORS], 'RF'"""
    cols = table.column_names
    front = [c for c in ['date', 'Mkt-RF'] if c in cols]
    back = [c for c in ['RF_AQR', 'RF_Q', 'RF'] if c in cols]
    fixed = set(front + back) #fix: don't define set in the loop! Now list comp is O(1)
    mid = [c for c in cols if c not in fixed]
    return table.select(front + mid + back)


def filter_table_by_date(table: pa.Table,
                         start: str | None,
                         end: str | None) -> pa.Table:
    """Slices a table to a start-end range."""
    if start is None and end is None:
        return table

    # target the date col and its type
    date_col = 'date' if 'date' in table.column_names else table.column_names[0]
    target_type = table.schema.field(date_col).type
    expr = pc.field(date_col)

    mask = None
    if start:
        mask = (expr >= pc.scalar(start).cast(target_type))

    if end:
        end_mask = (expr <= pc.scalar(end).cast(target_type))
        mask = (mask & end_mask) if mask is not None else end_mask

    return table.filter(mask)


def _validate_columns(table: pa.Table, names: str | list[str] | None) -> list[str]:
    """Private helper: Checks if columns exist in a pyarrow Table by name.

    Returns the list of validated factors or raises if missing.
    """
    if table.num_rows == 0:
        raise RuntimeError("Table is empty.")

    if not names:
        return []

    input_list = [names] if isinstance(names, str) else list(names)

    cols = set(table.column_names)

    missing = [f for f in input_list if f not in cols]
    if missing:
        raise ValueError(f"Columns not found in Table: {missing}")

    return input_list


def select_table_columns(table: pa.Table, factors: str | list[str]) -> pa.Table:
    """Helper: constructs a table from a pa.Table by selecting cols by name.

    'date' is col 0, and if passed in 'factors', is depduplicated.
    Used by base FactorModel's __getitem__, _extract_as_table and .extract()

    Case insensitive.
    """
    req_cols = [factors] if isinstance(factors, str) else list(factors)

    name_map = {col.lower(): col for col in table.column_names}

    valid_names = [name_map.get(f.lower(), f) for f in req_cols]
    valid_cols = _validate_columns(table, valid_names)

    factor_cols = [f for f in valid_cols if f.lower() != 'date']
    if not factor_cols:
        raise ValueError("Extraction must include at least one factor.")   ### TEST THIS

    # ['date'] + [verified factors minus date]
    selection = list(dict.fromkeys(['date'] + factor_cols))

    return table.select(selection).combine_chunks()


def _format_for_preview(val, col_name, precision=6):
    """Private helper for print_table_preview output. Rounds for display converts None to NaNs."""
    if val is None or (isinstance(val, float) and val != val): 
        return "NaN"

    if isinstance(val, (float, int)) and col_name != "date":
        prec = 4 if col_name.upper() in {'RF', 'RF_AQR'} else precision
        return f"{val:.{prec}f}"

    return str(val)



def print_table_preview(table, n_rows=6, n_cols=6) -> str | None:
    """Formats a str preview of a PyArrow table like a Pandas DataFrame.

    - If rows or columns exceed (n*2 + 2) then table is truncated.
    - Hardcoded 'date' on its own line. 
    - Prints "[...]" in place of truncated rows/cols.
    - Prints a footer of [rows x cols, size]
   
    Args:
        table (pa.Table): The table to preview.
        n_rows: Number of rows to show at head and tail.
        n_cols: Number of columns to show at start and end. Excludes date.
    """
    # TODO: way too much going on. Needs refactor. Own file.
    total_rows = table.num_rows
    if total_rows == 0:
        log.warning("Empty Table")
        return None
    
    precision = getattr(table, '_precision', 6) # model's own precision property
    all_columns = table.column_names
    total_cols = len(all_columns)
    
    # Makes sure at least 3 rows are hidden
    is_row_gap = total_rows > (n_rows * 2) + 2
    if is_row_gap:
        head = table.slice(0, n_rows).to_pylist()
        tail = table.slice(total_rows - n_rows, n_rows).to_pylist()
    else:
        head = table.to_pylist()
        tail = []

    # (total_cols-1) because all_columns[0] = 'date'
    data_cols_count = total_cols - 1
    is_col_gap = data_cols_count > (n_cols * 2) + 2
    
    if is_col_gap:
        # date + first N + gap + last N
        display_cols = (
            [all_columns[0]] + 
            all_columns[1 : n_cols + 1] + 
            ["[...]"] + 
            all_columns[-n_cols:]
        )
    else:
        display_cols = all_columns

    col_widths = {}
    for col in display_cols:
        if col == "[...]":
            col_widths[col] = 7
        else:
            # Checks length of header and the formatted values in head/tail
            header_len = len(col)
            max_val_len = max(
                (len(_format_for_preview(r[col], col, precision)) for r in head + tail), 
                default=0
            )
            col_widths[col] = max(header_len, max_val_len) + 2

    output = []
    
    # headers
    # 'date' is on a second line  (mimicing pandas)
    header_row = ""
    for col in display_cols[1:]:
        if col == "[...]":
            header_row += "[...]".center(col_widths[col])
        else:
            header_row += col.rjust(col_widths[col])

    output.append(f"{' ' * col_widths[all_columns[0]]}{header_row}")
    output.append(all_columns[0]) # word 'date'


    # data
    display_rows = head + ([None] if is_row_gap else []) + tail
    for row in display_rows:
        if row is None:
            output.append(" [...]")
            continue

        # date col format
        line = _format_for_preview(row[all_columns[0]], 'date', precision).ljust(col_widths[all_columns[0]])
        
        # format all other data cols
        for col in display_cols[1:]:
            if col == "[...]":
                line += "[...]".center(col_widths[col])
            else:
                val = _format_for_preview(row[col], col, precision)
                line += val.rjust(col_widths[col])
        output.append(line)

    # footer: [X rows x Y columns, Z kb]
    buf_size = table.get_total_buffer_size()
    output.append(f"\n[{total_rows} rows x {total_cols} columns, {buf_size / 1024:.1f} kb]")
    
    return "\n".join(output)



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import logging
from datetime import datetime
#import warnings
from io import BytesIO
from pathlib import Path
from types import MappingProxyType
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.csv as pv
from getfactormodels.utils.date_utils import offset_period_eom
from getfactormodels.utils.http_client import _HttpClient

log = logging.getLogger(__name__) #TODO: consistent logging.

"""Model utils and I/O utils."""

_MODEL_ALIASES = {
    "3": ["3", "ff3", "famafrench3"],
    "4": ["4", "ff4", "carhart", "car"],
    "5": ["5", "ff5", "famafrench5"],
    "6": ["6", "ff6", "famafrench6"],
    "Q": ["q", "qfactors", "q-factors", "q_factors", "q5", "hmxz"],
    "Qclassic": ["q4", "qclassic", "q-classic", "q_classic", "classic_q"],
    "HMLDevil": ["hmld", "hmldevil", "hml_devil", "devil"],
    "QMJ": ["qmj", "quality", "qualityminusjunk"],
    "BAB": ["bab", "betting", "bettingainstbeta"],
    "Mispricing": ["mispricing", "mis", "misp"],
    "Liquidity": ["liq", "liquidity"],
    "ICR": ["icr", "intermediary", "hkm"],
    "DHS": ["dhs", "behavioural", "behaviour"],
    "BarillasShanken": ["bs", "bs6", "barillasshanken", "barillas-shanken"],
    "VME": ["vme", "valmom", "valueandmomentumeverywhere"],
    "AQR6": ["aqr6", "aqr", "aqrfactors"],
    "HighIncomeCCAPM": ["hccapm", 'hcapm', 'hc-capm'],
    "ConditionalCAPM": ["jwcapm", "plcapm", "jwccapm", "plccapm", "ccapm"],
}

_LOOKUP = MappingProxyType({
    alias: key 
    for key, aliases in _MODEL_ALIASES.items() 
    for alias in aliases
})

def _get_model_key(user_input: str | int) -> str:
    """Converts user input (e.g. 'ff3', 'hmld') to the model key.
    
    >>> _get_model_key('ff3')
    '3'
    >>> _get_model_key('Bab')
    'BAB'
    """
    if not user_input:
        return ""
        
    val = str(user_input).lower().strip()
    
    return _LOOKUP.get(val, val)


def _prepare_filepath(filepath: str | Path | None, filename: str) -> Path:
    if filepath is None:
        return Path.cwd() / filename

    is_explicit_dir = str(filepath).endswith(("/", "\\"))
    user_path = Path(filepath).expanduser()

    if is_explicit_dir or user_path.is_dir():
        final_path = user_path / filename
    else:
        if not user_path.suffix:
            user_path = user_path.with_suffix(".csv")
        final_path = user_path
    
    # make sure it exists!
    final_path.parent.mkdir(parents=True, exist_ok=True)
    return final_path


def _generate_filename(model: 'FactorModel') -> str: # type: ignore [reportUndefinedVariable]   TODO: FIXME
    """Private helper: create filename using metadata from a model instance.
    
    Used if directory is provided, or just an extension, intended to be used 
    for -o flag with no param.
    """
    # 3 to "ff3", FF models only ones that accept int.
    raw_name = getattr(model, 'model', model.__class__.__name__.replace('Factors', ''))
    model_name = f"ff{raw_name}" if str(raw_name).isdigit() else raw_name

    freq = model.frequency
    region = getattr(model, 'region', None)

    if hasattr(model, 'data') and model.data.num_rows > 0:
        date_col = model.data.column(0)

        # pc.min/max give scalar, as_py makes scalar dt
        start_dt = pc.min(date_col).as_py()
        end_dt = pc.max(date_col).as_py()

        date_str = f"{start_dt.strftime('%Y%m%d')}-{end_dt.strftime('%Y%m%d')}"
    else:
        date_str = f"no_data_{datetime.now().strftime('%Y%m%d_%H%M%S')}"

    parts = [model_name, freq, region, date_str]
    filename = "_".join(str(p).lower() for p in parts if p)

    return f"{filename}.csv"
    

def _save_to_file(table: pa.Table, filepath: str | Path, model_instance=None):
    """Private helper: save a table to file.
    
    - Uses pyarrow, falls back to pandas to write .pkl.
    """
    _name = _generate_filename(model_instance) if model_instance else "factors.csv"
    
    full_path = _prepare_filepath(filepath, _name)
    
    ext = full_path.suffix.lower()
    table = table.combine_chunks()

    try:
        if ext == '.parquet':
            import pyarrow.parquet as pq
            pq.write_table(table, full_path, compression='snappy')
            
        elif ext in ('.feather', '.arrow', '.ipc'):
            import pyarrow.feather as pf
            pf.write_feather(table, full_path, compression='uncompressed')

        elif ext == '.csv':
            write_options = pv.WriteOptions(include_header=True)
            pv.write_csv(table, full_path, write_options=write_options)

        elif ext == '.txt':
            write_options = pv.WriteOptions(delimiter='\t')
            pv.write_csv(table, full_path, write_options=write_options)

        elif ext in ('.md', '.markdown'):
            with open(full_path, 'w', encoding='utf-8') as f:
                for line in _stream_table_to_md(table, precision = 6):
                    f.write(line + '\n')

        elif ext == '.pkl':
            df = table.to_pandas()
            df.to_pickle(full_path)
        else:
            supported = ['.parquet', '.feather', '.csv', '.txt', '.pkl', '.md']
            raise ValueError(f"Extension '{ext}' not supported. Options: {supported}")

    except Exception as e:
        raise OSError(f"Failed to write {ext} file to {full_path}: {e}") from e


def _stream_table_to_md(table: pa.Table, precision: int = 4):
    """Generator that yields markdown rows.
    
    - Reduces memory usage for writing larger tables.
    """
    try:
        yield "| " + " | ".join(table.column_names) + " |"
        yield "| " + " | ".join(["----"] * len(table.column_names)) + " |"

        str_columns = []
        rfs = {'RF', 'R_F', 'AQR_RF', 'RF_AQR'} 

        # clean floats, col presentation
        for name in table.column_names:
            col = table.column(name)  
            if pa.types.is_floating(col.type):
                prec = 4 if name.upper() in rfs else precision
                data = col.to_pylist()
                # f-strings to force decimals.
                # fix: prepend positives with an empty space.
                # - a space bef the f-string dot: space for positive, '-' for neg.
                # fix: alignment when col is all positives, rm the double spaces.
                contains_negs = pc.any(pc.less(col.fill_null(0), 0)).as_py()
                fmt = f" .{prec}f" if contains_negs else f".{prec}f"
                clean_col = [f"{x:{fmt}}" if x is not None else "" for x in data]
            else:
                clean_col = col.cast(pa.string()).to_pylist()
            str_columns.append(clean_col)

        for row in zip(*str_columns):
            yield "| " + " | ".join(row) + " |"
    except Exception as e:
                    msg = f"Failed to write markdown to {full_path}"
                    log.exception(msg) 
                    raise OSError(f"{msg}: {e}")


def read_from_fred(series_map: dict, frequency: str, client) -> pa.Table:
    """Internal utility to download and align FRED time-series data into a pa.Table.

    Args:
        series: dict of series_id, col renames
        frequency: fred standard codes for frequency (y, q, m).
        client: Optional _HttpClient. If None, manages its own lifecycle.

    """
    freq_map = {'y': 'Annual', 'm': 'Monthly', 'q': 'Quarterly'}
    fred_freq = freq_map.get(frequency[0], 'Monthly')
    
    _client = client if client else _HttpClient()
    
    try:
        tables = []
        for s_id, col_name in series_map.items():
            url = f"https://fred.stlouisfed.org/graph/fredgraph.csv?id={s_id}&fq={fred_freq}" #agg not needed yet
            data = _client.download(url)
            
            t = pv.read_csv(BytesIO(data))
            t = t.rename_columns(["date", col_name])
            
            t = t.cast(pa.schema([
                ('date', pa.date32()),
                (col_name, pa.float64()),
            ]))
            tables.append(t)

        result = tables[0]
        for next_table in tables[1:]:
            result = result.join(next_table, keys="date")
        
        t = result.sort_by([("date", "ascending")])
        return offset_period_eom(t, frequency)

    finally:
        if client is None:
            _client.close()




# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any
from typing import Literal
import pyarrow as pa
from getfactormodels.utils.arrow_utils import (
    filter_table_by_date,
    print_table_preview,
    rearrange_columns,
    round_to_precision,
    select_table_columns,
)
from getfactormodels.utils.date_utils import (
    _validate_input_date,
    validate_date_range,
)
from getfactormodels.utils.http_client import _HttpClient
from getfactormodels.utils.utils import _save_to_file

"""An abstract base class for model implementations.

- FactorModel: abstract base class. Provides common data handling, 
  caching, and date-filtering logic implemented by all factor models.
- RegionMixin: A mixin for models supporting international data.
- CompositeModel: base class for models built from other models.
- ModelCollection: for returning multiple models in a combined table.
"""
logger = logging.getLogger(__name__)


class FactorModel(ABC):
    """Abstract Base Class used by all factor model implementations."""
    def __init__(self, frequency: str | None = 'm',
                 start_date: str | None  = None,
                 end_date: str | None = None,
                 output_file: str | None = None,
                 cache_ttl: int = 86400,
                 **kwargs: Any):
        """Initialize the factor model instance.

        Args:
            frequency (str): the frequency of the data. Default: 'm'.
            start_date (str, opt): The start date YYYY[-MM-DD]
            end_date (str, opt): The end date YYYY[-MM-DD]
            output_file (str): optional path to save data to.
            cache_ttl (int): Cache time-to-live in seconds. Default: 86400.
            **kwargs: some models have additional params.
        """
        self.log = logger

        self._data: pa.Table | None = None
        self._start_date = None
        self._end_date = None
        self._frequency = None

        self.frequency = frequency 
        self.start_date = start_date 
        self.end_date = end_date
        self.output_file = output_file
        self.cache_ttl = cache_ttl

        self.copyright: str = ""  # NEW, TEST. fix: Carhart erroring with FF with copyright
        self._selected_factors: list[str] | None = None  # for eg drop/extract  # changing
        super().__init__()

    def __len__(self) -> int:
        """Length of the pa.Table after filtering."""
        return len(self.data)

    def __str__(self) -> str:
        if self.data is None:
            return self.__repr__()
        
       # table = self.data

        if isinstance(self, RegionMixin):
            region_label = f" ({self.region})"
        else:
            region_label = ""

        header = f"{self.__class__.__name__}{region_label}\n"
        return header + print_table_preview(self.data)

    def __repr__(self) -> str:
        params = []
        attrs = [
            'model', 'frequency', 'region', 'start_date', 'end_date',
            'country', 'classic', 'output_file',
        ]

        for attr in attrs:
            val = getattr(self, attr, None)

            if val is not None and val is not False and val != "":
                # format the str with quotes
                repr_val = f"'{val}'" if isinstance(val, str) else val
                params.append(f"{attr}={repr_val}")

        return f"{self.__class__.__name__}({', '.join(params)})"

    def __getitem__(self, key: str | list[str]) -> pa.Table:
        """Returns a pa.Table of date + selected factors."""
        return select_table_columns(self.data, key)

    def _repr_html_(self) -> str:
        """HTML repr for nice IPython/Jupyter outputs."""
        style = (
            "font-family: monospace; "
                "font-size: 0.9em; "
                "line-height: 1.4; "
                "background-color: transparent; "
                "white-space: pre; "
                "overflow-x: auto; "
                "display: block;"
        )

        # str(self) calls print_table_preview, wrap in <pre>
        return f'<pre style="{style}">{str(self)}</pre>'

    @property
    def start_date(self) -> str | None:
        return self._start_date
    @start_date.setter
    def start_date(self, value: Any):
        valid = _validate_input_date(value, is_end=False)
        self._start_date, self._end_date = validate_date_range(valid, self._end_date)

    @property
    def end_date(self) -> str | None:
        return self._end_date
    @end_date.setter
    def end_date(self, value: Any):
        valid = _validate_input_date(value, is_end=True)
        self._start_date, self._end_date = validate_date_range(self._start_date, valid)

    @property
    def frequency(self) -> str | None: 
        return self._frequency
    @frequency.setter
    def frequency(self, value: str | None):
        if value is None:
            self._frequency = None
            return

        val = value.lower()
        if val not in self._frequencies:
            raise ValueError(f"Invalid '{val}'. Options: {self._frequencies}") 

        if val != self._frequency:
            if self._frequency is not None:
                msg = f"Freq. changed from {self._frequency} to {val}." 
                self.log.info(msg)
            self._frequency = val
            self._data = None


    @property
    def data(self) -> pa.Table:
        """Returns a pa.Table with requested data (sliced)."""
        if self._data is None:
            self.load()   # Don't know if good idea
            
        table = self._data

        # drop/extract handled here.
        if self._selected_factors is not None:
            selection = list(dict.fromkeys(['date'] + self._selected_factors))
            table = table.select(selection)

        return filter_table_by_date(table, self._start_date, self._end_date)


    @property
    def shape(self) -> tuple[int, int]:
        """(rows, columns), like Pandas/Numpy."""
        return self.data.shape
   

    def join(self, other: "FactorModel") -> "ModelCollection":
        """Joins this model with another model or portfolio."""
        from getfactormodels.models.base import ModelCollection
        return ModelCollection(instances=[self, other])


    def extract(self, factor: str | list[str]) -> "FactorModel":   #Self
        """Select specific factors from the model. Str or list[str]."""
        table = select_table_columns(self.load(), factor)
        self._selected_factors = [f for f in table.column_names if f != 'date']
        return self


    def drop(self, factor: str | list[str]) -> "FactorModel": #Self
        """Remove specific factors from the model. Str or list[str]."""
        t_cols = self.load().column_names
        to_drop = {f.lower() for f in ([factor] if isinstance(factor, str) else factor)}

        # select: cols (lowercase) not in the to_drop set
        selection = [c for c in t_cols if c.lower() not in to_drop and c != 'date']

        if not selection:
            raise ValueError("Cannot drop all factors from the model.")

        self._selected_factors = selection
        return self


    def to_file(self, filepath: str | Path | None = None) -> None:
        """Save data to a file.

        Supports .parquet, .ipc, .feather, .csv, .txt, .pkl, .md
        
        Args:
            filepath (str | Path, optional): filepath to save data to.
        """
        target = filepath if filepath else self.output_file
        if not target:
            self.log.error("No filepath provided and no default output_file set.")
            return

        # self.data does get_table, selection, and date slicing.
        table = self.data

        if table.num_rows == 0:
            self.log.warning("No data available to save.")
            return

        _save_to_file(table, target, model_instance=self)


    def to_pandas(self) -> "pd.DataFrame":
        """Convert model to a pandas DataFrame.

        - Triggers the download if not loaded.
        """
        try:
            import pandas as pd  # check for if user has pd?
            df = self.data.to_pandas()
            if "date" in df.columns:
                df = df.set_index("date")
            return df
        except ImportError:
            raise ImportError("Requires Pandas. Try `pip install pandas`") from None


    def to_polars(self) -> "pl.DataFrame":
        """Convert model to a polars DataFrame.

        - Wrapper around Polars' `from_arrow()`. 
        - Triggers the download if not loaded.
        """
        try:
            import polars as pl
            return pl.from_arrow(self.data)
        except ImportError:
            raise ImportError("Requires Polars. Try `pip install polars`") from None
    # maybe
    #def preview(self, n: int = 4):
    #    """Prints the formatted table preview."""
    #    print(print_table_preview(self.data, n_rows=n))

    # RENAME: load, was _get_table
    def load(self, client: _HttpClient | None = None) -> pa.Table:
        """Trigger download or construction."""
        if self._data is not None:
            return self._data

        if hasattr(self, '_construct'):
            if client is None:
                with _HttpClient() as shared_client:
                    table = self._construct(shared_client)
            else:
                table = self._construct(client)
        else:
            raw_bytes = self._download(client=client)
            table = self._read(raw_bytes)

        # move this out probably ---------------------
        if "date" in table.column_names:
            table = table.sort_by([("date", "ascending")])

        table = round_to_precision(table, self._precision)

        # CompositeModel: if drop_null=True, drop
        if getattr(self, 'drop_null', False):
            table = table.drop_null()

        table.validate(full=True)
        self._data = rearrange_columns(table=table).combine_chunks()
        # -------------------------------------------
        return self


    def _download(self, client: _HttpClient | None = None) -> bytes | dict[str, bytes]:
        urls = self._get_url()
        self.log.info(f"Downloading from: {urls}")
        def _download_method(client: _HttpClient, urls: str | dict):
            if isinstance(urls, str):
                # A client was given. Using it.
                return client.download(urls, self.cache_ttl)
            return {k: client.download(v, self.cache_ttl) for k, v in urls.items()}
        try:
            if client:
                return _download_method(client, urls)
            with _HttpClient() as new_client:
                return _download_method(new_client, urls)
        except Exception as e:
            self.log.error(f"Download failed: {e}")
            raise RuntimeError(f"Could not retrieve data for {self.__class__.__name__}") from e


    # might move to utils
    def __dataframe__(self, *, nan_as_null: bool = False):
        """Dataframe interchange protocol support.

        Casts date32 to ns.

        Args:
            nan_as_null: converts NaN to null

        Examples:
            model = FamaFrenchFactors(model='3')
            df = model.to_pandas()

            import polars as pl
            df = pl.from_arrow(model.data)

            import pandas as pd
            df = pd.api.interchange.from_dataframe(model.data)
        """
        table = self.data

        if table.column_names[0] == "date":
            col_type = table.schema.field(0).type

        if pa.types.is_date(col_type):
            date_ns = table.column(0).cast(pa.timestamp("ns")).combine_chunks()
            table = table.set_column(0, "date", date_ns)

        return table.combine_chunks().__dataframe__(nan_as_null=nan_as_null, allow_copy=True)


    @property
    def _precision(self) -> int: return 8

    @property
    @abstractmethod
    def schema(self) -> pa.Schema:
        pass

    @property
    @abstractmethod
    def _frequencies(self) -> list[str]:
        pass

    @abstractmethod 
    def _get_url(self) -> str | dict[str, str]:
        """Build the unique data source URL."""
        pass

    @abstractmethod
    def _read(self, data: bytes) -> pa.Table:
        """Read bytes into a pa.Table."""
        pass


class CompositeModel(FactorModel):
    """Base for models constructed from other models."""
    def __init__(self, frequency: str = 'm', *, drop_null: bool = True, **kwargs):
        self.drop_null = drop_null
        super().__init__(frequency=frequency, **kwargs)

    @abstractmethod
    def _construct(self, client: _HttpClient) -> pa.Table:
        pass
    
    def _get_url(self) -> str:
        raise NotImplementedError("CompositeModel: no remote source.")
    
    def _read(self, data: bytes) -> pa.Table:
        raise NotImplementedError("CompositeModel: _read called on a composite models.")


class ModelCollection(CompositeModel):
    """A ModelCollection holds multiple models and/or portfolios of the same frequency."""
    def __init__(self, model_keys: list[str] | None = None, 
                 instances: list[FactorModel] | None = None, **kwargs):
        from getfactormodels.main import model
        
        if instances:
            self.instances = instances
            self.model_keys = [i.__class__.__name__ for i in instances]
        elif model_keys:
            self.model_keys = model_keys
            self.instances = [model(m, **kwargs) for m in model_keys]
        else:
            raise ValueError("model_keys or instances required.")
        super().__init__(**kwargs)

    def __str__(self) -> str:
        if self._data is None:
            return self.__repr__()

        region_label = ""
        regional_inst = next((i for i in self.instances if isinstance(i, RegionMixin)), None)
        if regional_inst:
            region_label = f" ({regional_inst.region})"

        models_str = ", ".join(f"'{k}'" for k in self.model_keys)
        header = f"{self.__class__.__name__}: {models_str} {region_label}\n"
        return header + print_table_preview(self.data)

    @property
    def _frequencies(self) -> list[str]:
        # the intersection of frequencies btw models
        freqs = set(self.instances[0]._frequencies)
        for inst in self.instances[1:]:
            freqs &= set(inst._frequencies)
        return list(freqs)

    @property
    def schema(self) -> pa.Schema:
        seen = {'date'}
        fields = [pa.field('date', pa.date32())]
        for inst in self.instances:
            for field in inst.schema:
                if field.name not in seen:
                    fields.append(field)
                    seen.add(field.name)
        return pa.schema(fields)
    

    def _construct(self, client: _HttpClient) -> pa.Table:
        # .load() all instances
        for inst in self.instances:
            inst.load(client=client)
        
        # first table is the base. Redundant recast date32 here...
        base_table = self.instances[0].data
        if base_table.column("date").type != pa.date32():
            base_table = base_table.set_column(0, "date", base_table.column("date").cast(pa.date32()))
        
        final_table = base_table

        for inst in self.instances[1:]:
            next_table = inst.data
            
            if next_table.column("date").type != pa.date32():
                next_table = next_table.set_column(0, "date", next_table.column("date").cast(pa.date32()))
            
            # overlaps are kept from the first table. TODO: only actual dupe cols, 
            # every factor needs a unique ID.
            overlaps = set(final_table.column_names) & set(next_table.column_names)
            overlaps.discard('date')

            uniq_cols = [c for c in next_table.column_names if c not in overlaps]
            if "date" not in uniq_cols:
                uniq_cols.insert(0, "date")

            table_to_join = next_table.select(uniq_cols)

            # Portfolio to FactorModel: inner. FactorModel to FactorModel: left outer
            is_portfolio_join = any(isinstance(i, PortfolioBase) for i in self.instances)
            join_type = "inner" if is_portfolio_join else "left outer"

            final_table = final_table.join(
                table_to_join,
                keys="date",
                join_type=join_type,
            )

        return final_table.sort_by([("date", "ascending")])


class RegionMixin:
    """Mixin for models that support international regions/countries."""
    #TODO: map properly!
    _aliases = {
        'uk': 'gbr', 'ger': 'deu', 'sg': 'sgp',
        'hk': 'hkg', 'nl': 'nld', 'fr': 'fra',

        # FF and AQR (not VME) - shared countries
        'usa': 'us', 'us': 'usa',
        'jpn': 'japan', 'japan': 'jpn',

        # FF and AQR (not VME) - shared regions : aggregate portfolios
        'ex-us': 'global ex usa',
        'north-america': 'north america', 'north america': 'north-america',

        # FF regions : AQR Aggregate Portfolios. 
        # Note: AQR's 'pacific' includes japan.
        #for VME
        'global_stocks': 'all_equities',
        
        # Aus
        'australia': 'aus', 'au': 'aus',
    }

    @property
    @abstractmethod
    def _regions(self) -> list[str]:
        pass

    @property
    def region(self) -> str:
        if hasattr(self, "_region") and self._region is not None:
            return self._region
        if not self._regions:
            return ""

        # set default: usa, else first region if no usa in model's regions... 
        if 'us' in self._regions:
            return 'us'
        if 'usa' in self._regions:
            return 'usa'
        return self._regions[0]
            
    @region.setter
    def region(self, value: str | None):
        if value is None:
            self._region = None
            return

        val = str(value).strip().lower()
        # Try a direct match, then alias, else keep (to fail)
        resolved = val if val in self._regions else self._aliases.get(val, val)
        
        if resolved not in self._regions:
            raise ValueError(f"Region '{value}' not supported for {self.__class__.__name__}. "
                             f"Available: {self._regions}")

        current = getattr(self, "_region", None)
        if current is not None and current != resolved:
            self._data = None
            if hasattr(self, "log"):
                self.log.info(f"Region changed to {resolved}. Cache reset.")

        self._region = resolved
    
    @classmethod
    def list_regions(cls) -> list[str]:
        """List available regions without instantiation."""
        if isinstance(cls._regions, property):
            # reach into the fget of the property
            return cls._regions.fget(cls) 
        return cls._regions


class PortfolioBase(FactorModel, ABC):
    """Base class for portfolio return data."""
    def __init__(self, frequency: str = 'm',
                 weights: Literal['vw', 'ew'] = 'vw',
                 *,
                 dividends: bool = True,
                 **kwargs):
        super().__init__(frequency=frequency, **kwargs)

        self.dividends = dividends
        self.weights = weights.lower()

        if self.weights not in ('vw', 'ew'):
            msg = f'weights should be either "vw" or "ew", not {weights}'
            raise ValueError(msg)

    @property
    def _precision(self) -> int: return 6




#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import argparse
import logging
import sys
import textwrap
from importlib.metadata import PackageNotFoundError, version
import pyarrow.csv as pv 
import os 
from pathlib import Path
from getfactormodels.utils.utils import _generate_filename

log = logging.getLogger("getfactormodels")

# TODO: clean up cli. Especially help. Add list models.

def _get_version():
    # Avoids importing __init__ for the ver.
    try: return version("getfactormodels")
    except PackageNotFoundError: return "unknown"


def _cli_list_regions():
    """Helper to display regions and exit."""
    from getfactormodels.models.aqr_models import _AQRModel
    from getfactormodels.models.fama_french import FamaFrenchFactors
    
    print(f"\nFAMA-FRENCH MODELS:\n  {textwrap.fill(', '.join(FamaFrenchFactors.list_regions()), width=70)}")
    print(f"\nAQR MODELS:\n  {textwrap.fill(', '.join(_AQRModel.list_regions()), width=70)}")
    print("\n  Note: accepts aliases 'us', 'jpn', 'uk', and 'ger'.")
    sys.exit(0)


def parse_args() -> argparse.Namespace:
    """CLI arg parser for getfactormodels."""
    parser = argparse.ArgumentParser(
        prog='getfactormodels',
        description='Download datasets for various factor models.',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='''Example usage:
    getfactormodels --model ff3 --frequency m --start 2000-01-01 --end 2010
    getfactormodels -m 5 -f m --extract SMB RF -o '~/file.csv'
    getfactormodels -m ff6 --drop 'RF'
    getfactormodels -m hml_devil --region jpn
    getfactormodels -p industry 30
    getfactormodels --model ff3 liq --portfolio op bm 
    getfactormodels -m ff5 liq -p beta -n 10 
        ''', 
    )

    parser.add_argument('--ver', '--version', action='version', version=f'getfactormodels {_get_version()}')
    parser.add_argument('-q', '--quiet', action='store_true', help='Suppress output to console.')
    parser.add_argument('-v', '--verbose', action='store_true', help="verbose output (set log to debug)")
    
    parser.add_argument('-m', '--model', nargs="+", metavar="MODEL", 
                        help="The model/s to use, e.g., 'liquidity', 'icr', "
                        "'ff3'. Accepts ints for Fama-French models, 3, 4, 5, 6.")
                        # TODO: 'see all models with --list models'
    
    parser.add_argument('-f', '--frequency', type=str, default='m',
                        choices=['d', 'w', 'w2w', 'm', 'q', 'y'], metavar="FREQ",
                        help="Data frequency (default: 'm'). Note: 'w2w' (Wed-to-Wed) is "
                        "only available for q-factors.")
    
    parser.add_argument('-s', '--start', required=False, metavar="YYYY-MM-DD", 
                        help='the start date.')
    
    parser.add_argument('-e', '--end', required=False, metavar="YYYY-MM-DD", 
                        help='the end date.')

    parser.add_argument('-o', '--output', type=str, required=False, default=None, metavar="PATH",
                        help='filename/filepath to save the data to.')

    parser.add_argument('-d', '--drop', nargs='+', metavar='FACTOR', 
                        help="drop specific factor(s) from a model.")

    parser.add_argument('-x', '--extract', nargs='+', metavar="FACTOR",
                        help='extract specific factor(s) from a model.')

    parser.add_argument('-r', '--region', dest='region', metavar='REGION',
                        help="Region or country code for AQR/FF models. "
                        "Use `--list-regions` to see all valid regions.")
    
    parser.add_argument('--list-regions', action='store_true', 
                        help="show all supported regions and exit")
    
    # Portfolio Options
    port_group = parser.add_argument_group('Portfolio Options')
    #formed on
    port_group.add_argument('-p', '--portfolio', '--on', '--by', 
                            dest='formed_on', nargs='+', metavar='FACTOR',
                            help="Factors to sort on (e.g., size, bm, inv) or 'industry'.")
    # sort
    port_group.add_argument('-n', '--sort', '--count', dest='sort', metavar='SORT',
                            help="Number of portfolios or grid (e.g., 10, 5x5, 2x3).")
    port_group.add_argument('-I', '--industry', type=int, dest='ind_count',
                            help="Shortcut for Fama-French industry portfolios (e.g., -I 12).")
    port_group.add_argument('-W', '--weights', choices=['vw', 'ew'], default='vw',
                            help="Weighting scheme (default: vw).")
    port_group.add_argument('--src', '--source', default='ff', choices=['ff', 'q'],
                            help="Data source: 'ff' (Fama-French) or 'q' (Q-factor/HXZ).")
    
    parser.set_defaults(industry=None)

    args = parser.parse_args()

    if args.formed_on:
        args.formed_on = [item.strip().lower() for s in args.formed_on for item in s.split(',')]
        
        # -p industry n
        if args.formed_on[0] in ['industry', 'ind']:
            if len(args.formed_on) > 1:
                args.ind_count = int(args.formed_on[1])
            else:
                args.ind_count = 12 # Default
            args.formed_on = None # It's an industry sort, not a factor sort
            args.sort = None
    
    #-p, or -I
    if args.ind_count:
        args.industry = args.ind_count
        args.sort = None
        args.formed_on = None
    else:
        args.industry = None

    args.is_portfolio = bool(args.industry or args.formed_on)

    return args


# From main.py
def _cli():
    from getfactormodels.main import model, portfolio
    args = parse_args()

    if args.list_regions:
        _cli_list_regions()
        sys.exit(0)

    try:
        rhs, lhs = None, None
        
        if args.model:
            rhs = model(
                model=args.model, 
                frequency=args.frequency, 
                start_date=args.start, 
                end_date=args.end, 
                region=args.region
            )

        if args.is_portfolio:
            lhs = portfolio(
                source=args.src, 
                industry=args.industry,
                formed_on=args.formed_on,
                sort=args.sort,  # -n flag
                weights=args.weights,
                frequency=args.frequency,
                start_date=args.start,
                end_date=args.end
            )

        if rhs and lhs:
            rhs.load()
            lhs.load()
            # fix: sort after join! (eg, -m misp ff3 -p 2x3 -b size op wasn't returning full table)
            _table = rhs.data.join(lhs.data, keys="date", join_type="inner").sort_by("date")
            # A FactorModel object is needed (for to_file/extract/drop etc.),
            # this uses the RHS instance, then updates its _data property.
            model_obj = rhs
            model_obj._data = _table
        else:
            model_obj = rhs or lhs
            if model_obj:
                model_obj.load()
            if not len(model_obj):
                log.error("No data returned.")
                sys.exit(1)

    except (ValueError, RuntimeError) as e:
        print(f"ERROR: {e}", file=sys.stderr)
        sys.exit(1)

    if args.extract:
        model_obj.extract(args.extract)
    elif args.drop:
        model_obj.drop(args.drop)

    if args.output:
        model_obj.to_file(args.output)

        if not args.quiet:
            actual_path = Path(args.output).expanduser()
            if actual_path.is_dir():
                actual_path = actual_path / _generate_filename(model_obj)

            print(f"Data saved to: {actual_path.resolve()}", file=sys.stderr)

    # fix: '%%bash' commands are run in a subprocess (output was entire table)
    nb_env = 'ipykernel' in sys.modules or 'JPY_PARENT_PID' in os.environ

    if not sys.stdout.isatty() and not nb_env: # piped
        # model_obj.data's been filtered. Write csv stream of it:
        pv.write_csv(model_obj.data, sys.stdout.buffer)

    else:
        # we're interactive/IPython: print preview of table to stderr. 
        # uses the model_obj's __str__ (which prints the Table preview)
        if not args.quiet:
            sys.stderr.write(f"{str(model_obj)}\n")



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import pyarrow as pa
from getfactormodels.utils.arrow_utils import select_table_columns
from getfactormodels.utils.http_client import _HttpClient
from .aqr_models import HMLDevilFactors
from .base import CompositeModel
from .fama_french import FamaFrenchFactors
from .q_factors import QFactors


class BarillasShankenFactors(CompositeModel):
    """Download the Barillas-Shanken 6 Factor Model (2018).

    Combines: Mkt-RF, SMB (FF), IA, ROE (Hou-Xue-Zhang), 
    and UMD, HML_Devil (AQR).

    References:
        F. Barillas and J. Shanken, 2018. Comparing Asset Pricing Models. 
        Journal of Finance, vol. 73, no. 2, pp. 715–754.

    """
    def _construct(self, client: _HttpClient) -> pa.Table:
        ff = FamaFrenchFactors(model='3', frequency=self.frequency).load(client=client)
        q = QFactors(frequency=self.frequency).load(client=client)
        aqr = HMLDevilFactors(frequency=self.frequency).load(client=client)

        mkt_smb = select_table_columns(ff.data, ['Mkt-RF', 'SMB']) # might switch to using aqr's data (SMB = SMB FF)
        ia_roe = select_table_columns(q.data, ['R_IA', 'R_ROE'])
        umd_hml = select_table_columns(aqr.data, ['UMD', 'HML_Devil', 'RF_AQR'])

        table = (
            mkt_smb.join(ia_roe, keys='date')
            .join(umd_hml, keys='date')
        )
        # TODO: helper util to drop NaNs only if continuous from the edges
        return table.select(self.schema.names).cast(self.schema)

    @property
    def _frequencies(self) -> list[str]: return ["d", "m"]
    
    @property
    def _precision(self) -> int: return 8
    
    @property
    def schema(self) -> pa.Schema:
        return pa.schema([
            ('date', pa.string()),
            ('Mkt-RF', pa.float64()),
            ('SMB', pa.float64()),
            ('HML_Devil', pa.float64()),
            ('R_IA', pa.float64()),
            ('R_ROE', pa.float64()),
            ('UMD', pa.float64()),
            ('RF_AQR', pa.float64()),
        ])



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
from getfactormodels import models as factor_models
from getfactormodels.models.base import FactorModel, RegionMixin
from getfactormodels.utils.utils import _get_model_key
from getfactormodels.utils.cli import _cli
import warnings 

log = logging.getLogger("getfactormodels")


def portfolio(
    source: str = 'ff',
    formed_on: str | list[str] = 'size',
    sort: str | int | None = None,
    industry: int | None = None,
    weights: str = 'vw',
    frequency: str = 'm',
    start_date: str | None = None,
    end_date: str | None = None,
    **kwargs
):
    """Download portfolio return data.

    * Currently supports Fama-French sorts and industry portfolios,
    US only.

    Args:
        source: Data source identifier (e.g., 'ff', 'q').
        formed_on: Factor(s) to sort on (e.g., 'size', 'bm').
        sort: 'decile', '5x5', also accepts integers (10, 25, etc.)
        industry: Number of industry portfolios.
        weights: Weighting scheme ('vw' or 'ew').
        frequency: Data frequency ('d', 'm', 'y').
        start_date: Optional start date YYYY-[MM-DD].
        end_date: Optional end date YYYY[-MM-DD].
    """
    source = source.lower()
    
    params = {
        "formed_on": formed_on,
        "sort": sort,
        "industry": industry,
        "weights": weights,
        "frequency": frequency,
        "start_date": start_date,
        "end_date": end_date,
        **kwargs
    }

    if source in ['ff', 'famafrench']:
        from getfactormodels.models.fama_french import _get_ff_portfolios
        return _get_ff_portfolios(**params)

    raise ValueError(f"Portfolio source '{source}' not recognized.")


def model(
    model: str | int | list[str | int] = 3,
    region: str = 'usa',
    frequency: str = 'm',
    start_date: str | None = None,
    end_date: str | None = None,
    **kwargs
) -> FactorModel:   #Self 
    """Download factor model data.
    
    Args:
        model (str | list[str]): Model identifier.
        region: Geographical region (e.g., 'usa', 'developed').
        frequency: Data frequency ('d', 'w', 'm', 'y').
        start_date: Optional start date (YYYY-MM-DD).
        end_date: Optional end date (YYYY-MM-DD).
    """
    if isinstance(model, list):
        if len(model) == 1:
            model = model[0]
        else:
            from getfactormodels.models.base import ModelCollection
            return ModelCollection(
                model_keys=model, 
                region=region, 
                frequency=frequency, 
                start_date=start_date, 
                end_date=end_date, 
                **kwargs
            )

    model_key = _get_model_key(model)

    model_class_map = {
        "3": "FamaFrenchFactors",
        "5": "FamaFrenchFactors",
        "6": "FamaFrenchFactors",
        "4": "CarhartFactors",
        "Qclassic": "QFactors",
        "HighIncomeCCAPM": "HighIncomeCCAPM",
        "ConditionalCAPM": "ConditionalCAPM",
    }

    class_name = model_class_map.get(model_key, f"{model_key}Factors")
    factor_class = getattr(factor_models, class_name, None)

    is_regional = issubclass(factor_class, RegionMixin)
    if not is_regional:
        if 'region' in kwargs or region not in [None, 'usa']: 
            raise ValueError(f"Model '{class_name}' is does not accept a region argument.")
    else:
        kwargs['region'] = region

    return factor_class(frequency=frequency, **kwargs)


def get_factors(*args, **kwargs): #noqa
    """DEPRECATED: Use `model()` instead.

    This function will be removed in a future release.
    """
    warnings.warn(
        "get_factors() is deprecated and will be removed in a future version. "
            "Please use model() for factor data or portfolio() for return data.",
        FutureWarning,
        stacklevel=2
    )
    return model(*args, **kwargs)


def main():
    _cli()

if __name__ == "__main__":
    main()



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
from io import BytesIO
import pyarrow as pa
from python_calamine import CalamineWorkbook
from getfactormodels.models.base import FactorModel
from getfactormodels.models.fama_french import FamaFrenchFactors
from getfactormodels.utils.arrow_utils import (
    scale_to_decimal,
    select_table_columns,
)
from getfactormodels.utils.date_utils import offset_period_eom

#WIP NOT WORKING

class STVFactors(FactorModel):
    """Download the STV of Ung, Gebka and Anderson (2024).

    Data from Dec 1968 - Dec 2023. 

    This is the 'updated' calculation, "using the latest 
    available data (31st May 2024) from Professor Wurgler's website."

    Not implemented yet: accessing the 2016, and 2022 calculations. TODO

    References:
        Sze Nie Ung, Bartosz Gebka & Robert D. J. Anderson, 2024. 
        An enhanced investor sentiment index*. The European 
        Journal of Finance, 30:8, 827-864.

    Source:
        Ung, Gebka and Anderson, University of Manchester Figshare, 
        26 Mar 2025, DOI: 10.48420/28445081.v1, 
        https://figshare.manchester.ac.uk/articles/dataset/Enhanced_Investor_Sentiment_Index_STV_/28445081/1 Dataset
    """
    @property
    def _frequencies(self) -> list[str]:
        return ["m"]

    @property
    def _precision(self) -> int: 
        return 10  #11 is bad, but there's 14 undecimalized. [TODO: new system to enforce precision...]

    @property
    def schema(self) -> pa.Schema:
        schema = [ 
            ("Date", pa.string()),
            ("Enhanced Investor Sentiment Index (STV)\nDec 1968 - Dec 2023", pa.float64()), 
        ]
        return pa.schema(schema)

    def _get_url(self):
        return "https://figshare.manchester.ac.uk/ndownloader/files/52498529"

    # TODO: make reader a func xlsx models can use (maybe not AQR)
    def _read(self, data: bytes) -> pa.Table:
        workbook = CalamineWorkbook.from_filelike(BytesIO(data))
        
        rows = workbook.get_sheet_by_index(2).to_python()  # just 3rd sheet for now (2024 calc)
        if not rows: 
            return None

        headers = rows[0]
        data_rows = rows[1:]

        table_data = {}
        for i, header in enumerate(headers):
            clean_header = str(header).strip().replace('\n', ' ')

            column_values = []
            for row in data_rows:
                val = row[i] if i < len(row) else None
                if val == "" or val == "None":
                    val = None
                column_values.append(val)
            table_data[clean_header] = column_values

        table = pa.Table.from_pydict(table_data)
        table = table.rename_columns(['date', 'STV'])

        table = offset_period_eom(table, self.frequency) 
        table = scale_to_decimal(table)
        table.validate(full=True)

        # Make func. Can't get mkt-rf and RF like this. Precision messing up too. Just RF for now...
        _ff = FamaFrenchFactors(model='3', frequency=self.frequency).load()
        _ff = select_table_columns(_ff.data, ['RF']) # RF's ok because of the hard precision on it

        return table.join(_ff, keys="date", join_type="inner").combine_chunks()



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import pyarrow as pa
import pyarrow.compute as pc
from getfactormodels.models.base import CompositeModel
from getfactormodels.utils.arrow_utils import scale_to_decimal
from getfactormodels.utils.date_utils import offset_period_eom
from getfactormodels.utils.utils import read_from_fred

### SCRATCH

#### AUS REGIONAL VERSION OF PremiumLabourCCAPM 

# might rename PremiumLabourCCAPM
class ConditionalCAPM(CompositeModel):
    @property
    def _frequencies(self) -> list[str]:
        return ['y']

    @property
    def schema(self) -> pa.Schema:
        return pa.schema([
            ('date', pa.int32()),
            ('LBR', pa.float64()),
            ('PREM', pa.float64()),
            ('Mkt-RF', pa.float64()),
        ])

    def _construct(self, client) -> pa.Table:
        data_series = {
            "LCWRTT01AUQ661N": "income", #Labor Compensation: Wage Rate: All Activities: Hourly for Australia (LCWRTT01AUQ661N) use ABS...
            "POPTOTAUA647NWDB": "pop",   #Population, Total for Australia (POPTOTAUA647NWDB), note: use ABS.
            "IR3TBB01AUM156N": "sty",   # IR3TBB01AUM156N, 3 month/90 days (TODO: do it properly with RBA)
            "IRLTLT01AUM156N": "lty",   # Interest Rates: Long-Term Government Bond Yields: 10-Year: Main (Including Benchmark) for Australia (IRLTLT01AUM156N)
        }

        fred_table = read_from_fred(series=data_series, 
                                    frequency=self.frequency, 
                                    aggregate='eop', 
                                    client=client)

        data_table = offset_period_eom(fred_table, self.frequency) #important: EOM offset before calc

        jw_table = self._jw_calc(data_table)

        # PREM to decimal
        prem_scaled = scale_to_decimal(jw_table.select(['PREM']))
        table = jw_table.drop(['PREM']).append_column('PREM', prem_scaled.column(0))

        table.validate()
        return table.combine_chunks()

        #ff = FamaFrenchFactors(model='3', frequency=self.frequency).load(client=client)
        #mkt = select_table_columns(ff.data, ['Mkt-RF', 'RF'])
        #return jw_table.join(mkt, keys="date", join_type="inner").combine_chunks()


    def _jw_calc(self, table: pa.Table) -> pa.Table:
        """J&W (1996) factor construction using FRED data.

        - LBR: Smooth growth in per-capita labor income.
        - PREM: Lagged yield spread (Baa - Aaa) as a conditioning variable.

        """ 
        # Per-capita labor income and Yield Spread
        l_pc = pc.divide(table.column("income"), table.column("pop"))
        spread = pc.subtract(table.column("sty"), table.column("lty"))

        n = table.num_rows
        out_len = n - 2

        # LBR: Growth at t (using t and t-1)
        l_t = l_pc.slice(1, out_len)
        l_t_minus_1 = l_pc.slice(0, out_len)
        r_labor = pc.subtract(pc.divide(l_t, l_t_minus_1), 1.0)

        # PREM: Spread at t-1
        prem_lagged = spread.slice(0, out_len)

        # dates aligned to period t
        dates = table.column("date").slice(2, out_len)

        return pa.Table.from_arrays(
            [dates, r_labor, prem_lagged],
            names=["date", "LBR", "PREM"],
        )



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import logging
from types import MappingProxyType
from typing import Final, Literal, TypeAlias

#import warnings

log = logging.getLogger(__name__) #TODO: consistent logging.

"""Model utils. Model map, aliases, lookup and typing."""

ModelKey: TypeAlias = Literal[
    "3", "4", "5", "6", "Q", "Qclassic", "HMLDevil", "QMJ", "BAB", 
    "Mispricing", "Liquidity", "ICR", "DHS", "BarillasShanken", "VME", "AQR6", "SimpleCAPM",
]

_MODEL_ALIASES: Final[dict[ModelKey, list[str]]] = {
    "3": ["3", "ff3", "famafrench3"],
    "4": ["4", "ff4", "carhart", "car"],
    "5": ["5", "ff5", "famafrench5"],
    "6": ["6", "ff6", "famafrench6"],
    "Q": ["q", "qfactors", "q-factors", "q_factors", "q5", "hmxz"],
    "Qclassic": ["q4", "qclassic", "q-classic", "q_classic", "classic_q"],
    "HMLDevil": ["hmld", "hmldevil", "hml_devil", "devil"],
    "QMJ": ["qmj", "quality", "qualityminusjunk"],
    "BAB": ["bab", "betting", "bettingainstbeta"],
    "Mispricing": ["mispricing", "mis", "misp"],
    "Liquidity": ["liq", "liquidity"],
    "ICR": ["icr", "intermediary", "hkm"],
    "DHS": ["dhs", "behavioural", "behaviour"],
    "BarillasShanken": ["bs", "bs6", "barillasshanken", "barillas-shanken"],
    "VME": ["vme", "valmom", "valueandmomentumeverywhere"],
    "AQR6": ["aqr6", "aqr", "aqrfactors"],
    "SimpleCAPM": ['capm'],
}

# Inverted Index: O(1) lookup
# This creates a flat dict: {'ff3': '3', 'carhart': '4', ...}
_LOOKUP = MappingProxyType({
    alias.lower(): key 
    for key, aliases in _MODEL_ALIASES.items() 
    for alias in [key] + aliases  # don't forget the key itself
})
_VALID_INPUTS: Final[frozenset[str]] = frozenset(_LOOKUP.keys())

def get_model_key(user_input: str | int) -> str:
    """Converts user input to the model key.

    >>> _get_model_key('3')
    '3'
    >>> _get_model_key('ff6')
    '6'
    >>> _get_model_key('q5')
    'Q'
    """
    if not user_input:
        return ""

    val = str(user_input).lower().strip()
    #return _LOOKUP.get(val, val)
    match = _LOOKUP.get(val, val)
    log.debug(f"Input '{user_input}' ('{val}'), resolved: '{match}'") 
    return match

def get_available_models() -> tuple[ModelKey, ...]:
    return tuple(_MODEL_ALIASES.keys())






# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
from .fama_french import FamaFrenchFactors


class CarhartFactors(FamaFrenchFactors):
    """Download and process the Carhart 4-Factor model data.
    
    This model extends the Fama-French 3-factor model to 4 factors, 
    adding a momentum factor (MOM).

    References:
        M. Carhart, 1997. On Persistence in Mutual Fund Performance.
        Journal of Finance, vol. 52, no. 1, pp. 57–82.
        ...

    """
    @property
    def _frequencies(self) -> list[str]: return ["d", "m", "y"]

    def __init__(self, frequency='m', region=None, **kwargs):
        """Initialize the Carhart 4-Factor model."""
        kwargs.pop('model', None)
        super().__init__(frequency=frequency,
                         model=4,   # enforce model for FF 
                         region=region,
                         **kwargs)



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import io
import zipfile
from typing_extensions import override
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.csv as pv
from getfactormodels.models.base import FactorModel, RegionMixin, PortfolioBase
from getfactormodels.utils.arrow_utils import (
    round_to_precision,
    scale_to_decimal,
)
from getfactormodels.utils.date_utils import offset_period_eom
from abc import ABC, abstractmethod

#TODO: break up _read_zip
class FamaFrenchFactors(FactorModel, RegionMixin):
    """Download Fama-French (and Carhart) factor models.

    Downloads the 3-factor (1993), 5-factor (2015), or 6-factor (2018)
    model of Fama & French, or Carhart's (1997) 4-factor model.

    Weekly data only available for the 3-factor model.

    Args:
        model (str | int): model to return. '3' '4' '5' '6' (default: '3')
        frequency (str, optional): frequency of the data. 'd' 'm' 'y' 'w'
            (default: 'm').
        start_date (str, optional): start date of the data. YYYY[-MM-DD].
        end_date (str, optional): end date of the data. YYYY[-MM-DD].
        cache_ttl (int, optional`): cached download time-to-live in seconds 
            (default: 86400).
        region (str, optional): return an international/emerging market 
            model.

    References:
    - E. F. Fama and K. R., 1993. French. Common risk factors in the returns 
      on stocks and bonds. Journal of Financial Economics, vol. 33, no. 1,
      pp. 3–56.
    - E. F. Fama and K. R. French, 2015. A five-factor asset pricing model. 
      Journal of Financial Economics, vol. 116, no. 1, pp. 1–22.
    - E. F. Fama and K. R. French, 2018. Choosing factors. Journal of 
      Financial Economics, vol. 128, no. 2, pp. 234–252, 2018.

    """
    # TODO: NaNs in FamaFrench models! check
    #Note: "-0.9999" should be NaNs! [TODO: FIXME]
    @property
    def _frequencies(self) -> list[str]: return ['d', 'w', 'm', 'y']
    
    @property 
    def _precision(self) -> int: return 6

    @property 
    def _regions(self) -> list[str]:
        return [
            'us', 'emerging', 'developed', 'ex-us', 'europe',
            'japan', 'asia-pacific-ex-japan', 'north-america', 
        ]

    def __init__(self, 
                 frequency: str = 'm', 
                 model: int | str = '3', 
                 region: str | None = 'us', 
                 **kwargs) -> None:
        """Initialize a Fama-French factor model."""
        self.model = str(model)
        super().__init__(frequency=frequency, model=model, **kwargs)
        
        self.region = region

        self._validate_ff_input()
    

    @property
    def schema(self) -> pa.Schema:
        """Fama-French schema for specific model/frequency/region."""
        cols = [("date", pa.string()), 
                ("Mkt-RF", pa.float64()), 
                ("SMB", pa.float64()), 
                ("HML", pa.float64())]
        
        # 5 and 6-factor models: RMW and CMA
        if self.model in ["5", "6"]:
            cols += [("RMW", pa.float64()), 
                     ("CMA", pa.float64())]

        # mom models (4/6): dev/emerging: WML. US 6 factor: UMD. US 4: MOM.
        if self.model in ["4", "6"]:
            is_intl = self.region not in ["us", None]
            mom_name = "WML" if is_intl else ("UMD" if self.model == "6" else "MOM")
            cols.append((mom_name, pa.float64()))
        cols.append(("RF", pa.float64()))
        
        return pa.schema(cols)

    @property
    def _mom_schema(self) -> pa.Schema:
        """Private helper: schema for momentum files. 'val' is placeholder."""
        return pa.schema([("date", pa.string()), 
                          ("val", pa.float64())])
    @property
    def model(self) -> str:
        return self._model
    @model.setter
    def model(self, value: int | str):
        val = str(value)
        if val not in ["3", "4", "5", "6"]:
            raise ValueError(f"Invalid model '{val}'. Options: 3, 4, 5, 6")
        
        if hasattr(self, '_model') and val != self._model:
            self.log.info(f"Model changed from {self._model} to {val}")
            self._data = None
            self._view = None
        self._model = val 

    @property
    def _ff_region_map(self) -> dict[str, str]:
        """Private: maps region input to region URL str"""
        return {
            'us': 'US',
            'emerging': 'Emerging',
            'developed': 'Developed',
            'ex-us': 'Developed_ex_US',
            'europe': 'Europe',
            'japan': 'Japan',
            'asia-pacific-ex-japan': 'Asia_Pacific_ex_Japan',
            'north-america': 'North_America',
        }

    def _validate_ff_input(self) -> None:
        """Validates input for Fama-French models for things getters/setters dont take care of."""
        if self.region != 'us' and self.frequency == 'w':
            raise ValueError(f"Weekly frequency not available for {self.region}.")

        if self.region == "emerging" and self.frequency != 'm':
            raise ValueError("Emerging markets only available in monthly.")

        if self.frequency == 'w' and self.model != "3":
            raise ValueError("Weekly Fama-French data is only available for the 3-factor model.")
 

    def _get_url(self) -> dict[str, str]:
        """Constructs the URLs for downloading Fama-French data."""
        base_url = "https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp"
        url_list = {}

        ff_url_name = self._ff_region_map.get(self.region)
        
        if ff_url_name == 'Emerging':
            # Emerging: only monthly 5-factors avail
            filename = "Emerging_5_Factors_CSV.zip"
        elif ff_url_name == 'US':
            _model = "F-F_Research_Data_5_Factors_2x3" if self.model in {"5", "6"} else "F-F_Research_Data_Factors"
            freq_map = {'d': '_daily', 'w': '_weekly'}
            suffix = freq_map.get(self.frequency, "")
            filename = f"{_model}{suffix}_CSV.zip"
        else:
            # International regions
            base_model = '3' if self.model in ['3', '4'] else '5'
            daily = '_Daily' if self.frequency == 'd' else ''
            filename = f"{ff_url_name}_{base_model}_Factors{daily}_CSV.zip"
            
        url_list['factors'] = f"{base_url}/{filename}"
        
        # Construct the momentum URL if needed.
        if self.model in ["4", "6"]:
            freq_suffix = "_daily" if self.frequency == 'd' else ""
            if self.region in ['us', None]:
                mom_filename = f"F-F_Momentum_Factor{freq_suffix}_CSV.zip"
            else:
                mom_filename = f"{self.region}_Mom_Factor{freq_suffix.title()}_CSV.zip"
            
            url_list['momentum'] = f"{base_url}/{mom_filename}"

        return url_list


    def _read_zip(self, _data: bytes, use_schema: pa.Schema = None) -> pa.Table:
        """Download and read CSV from zip file."""
        _schema = use_schema if use_schema else self.schema

        with zipfile.ZipFile(io.BytesIO(_data)) as z:
            filename = z.namelist()[0]
            with z.open(filename) as f:
                lines = f.read().decode('utf-8').splitlines()

        is_mom = 'momentum' in filename.lower() or '_mom_' in filename.lower()
        annual_marker = next((i for i, l in enumerate(lines) if "annual factors:" in l.lower()), None)

        if self.frequency == 'y':
            raw_content = lines[annual_marker + 1:] if annual_marker is not None else lines
        else:
            _start = 6 if self.region == 'emerging' else (12 if is_mom else 4)
            _stop = annual_marker if annual_marker else len(lines)
            raw_content = lines[_start:_stop]

        content = [",".join(_schema.names)]
        for line in raw_content:
            clean = line.strip()
            if not clean or "copyright" in clean.lower():
                if "copyright" in clean.lower(): self.copyright = clean
                continue
            
            # Use commas for split, not spaces!
            parts = [p.strip() for p in clean.split(',')]
            
            # only keeps lines that start with a date
            if parts and parts[0].isdigit():
                content.append(",".join(parts))

        return pv.read_csv(
            io.BytesIO("\n".join(content).encode('utf-8')),
            convert_options=pv.ConvertOptions(
                null_values=["-99.99", "-999", "-99.990", "-0.9999"],
                strings_can_be_null=True,
                column_types=_schema,
                include_columns=_schema.names,
            ),
        )   
    

    def _read(self, data: dict[str, bytes]) -> pa.Table:
        main_cols = [f for f in self.schema if f.name not in ["UMD", "MOM", "WML"]]
        
        # Emerging Markets: base file is 5 factors (even for models 3/4)
        if self.region == 'emerging' and self.model in ["3", "4"]:
            main_cols.insert(4, pa.field("RMW", pa.float64()))
            main_cols.insert(5, pa.field("CMA", pa.float64()))

        main_schema = pa.schema(main_cols)
        # data['factors']
        table = self._read_zip(data['factors'], use_schema=main_schema)

        if 'momentum' in data:
            mom_table = self._read_zip(data['momentum'], use_schema=self._mom_schema)
            
            # momentum column name based on schema
            mom_key = next(k for k in ["UMD", "MOM", "WML"] if k in self.schema.names)
            mom_table = mom_table.rename_columns(["date", mom_key])
            
            table = table.join(mom_table, keys="date", join_type="inner").combine_chunks()

        table = table.set_column(0, "date", table.column(0).cast(pa.string()))
        table = offset_period_eom(table, self.frequency)
        table = scale_to_decimal(table)
        
        table = round_to_precision(table, self._precision)

        return table.select(self.schema.names).combine_chunks()


# WIP ---------------------------------------------------------------------- #
# TODO: FactorModel to BaseModel.
class _FFPortfolioBase(PortfolioBase, ABC):
    """Base class for Fama-French portfolio return data."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abstractmethod
    def _find_table_start(self) -> str:
        """A pattern to identify the specific table's start."""
        ...

    def _get_ff_table(self, lines: list[str]) -> str:
        weight_ln = "Value" if self.weights == 'vw' else "Equal"
    
        # fix: US bivariate size/OP 2x3 sorts contain an error. Header line in the daily 
        # file is labelled 'Monthly' 
        is_size_op_2x3 = (
            hasattr(self, 'formed_on') and 
            frozenset(self.formed_on) == frozenset(['size', 'op']) and
            getattr(self, 'n_portfolios', None) in [6, "6"]
        )
        
        if is_size_op_2x3 and self.frequency == 'd':
            freq_label = "Monthly" 
        else:
            freq_label = "Annual" if self.frequency == 'y' else ("Daily" if self.frequency == 'd' else "Monthly")
        table_start = None
        for i, ln in enumerate(lines):
            if weight_ln in ln and "Returns" in ln and freq_label in ln:
                table_start = i
                break

        if table_start is None:
            raise ValueError(
                f"Could not find table header: `{weight_ln}`"
            )

        # Skip title line, find first non empty line
        header_idx = table_start + 1
        while header_idx < len(lines) and not lines[header_idx].strip():
            header_idx += 1

        if header_idx >= len(lines):
            raise ValueError("Couldn't find table header (found table title).")

        headers = ("date " + " ".join(lines[header_idx].split())).lower().replace("-", "_")
        cleaned = [headers]

        # Extract lines
        for line in lines[header_idx + 1:]:
            clean = line.strip()
            # Table's over if line doesn't start with a number
            if not clean or not clean[0].isdigit():
                break
            cleaned.append("  ".join(clean.split()))

        return "\n".join(cleaned)


    def _read(self, data: bytes) -> pa.Table:
        with zipfile.ZipFile(io.BytesIO(data)) as z:
            with z.open(z.namelist()[0]) as f:
                lines = f.read().decode('utf-8').splitlines()

        clean_csv = self._get_ff_table(lines)

        table = pv.read_csv(
            io.BytesIO(clean_csv.encode('utf-8')),
            convert_options=pv.ConvertOptions(
                null_values=["-99.99", "-999", "-99.9"],
                column_types={"date": pa.string()},
            ),
        )

        table = self._fix_ff_nulls(table)
        table = offset_period_eom(table, self.frequency)
        table = scale_to_decimal(table)

        renames = [n.strip().lower().replace(" ", "_").replace("-", "_") for n in table.column_names]
        table = table.rename_columns(renames)

        return table.combine_chunks()


    def _fix_ff_nulls(self, table: pa.Table) -> pa.Table:
        ff_nans = pa.array([-99.99, -999.0, -99.9], type=pa.float64())
        for i, name in enumerate(table.column_names):
            if name.lower() == "date": continue
            col = table.column(name)
            mask = pc.is_in(col, value_set=ff_nans)
            table = table.set_column(i, name, pc.if_else(mask, pa.scalar(None, type=col.type), col))
        return table


    @property
    def schema(self) -> pa.Schema:
        """Fama-French portfolio schema."""
        if self._data is not None:
            return self._data.schema
        return pa.schema([pa.field("date", pa.date32())])
    
    @property
    def _precision(self) -> int: return 6


class _FamaFrenchSorts(_FFPortfolioBase):
    @property
    def _frequencies(self) -> list[str]:
        return ['d', 'w', 'm', 'y']
    
    #@property 
    #def _regions(self) -> list[str]:
    #    return [
    #        'us', 'emerging', 'developed', 'ex-us', 'europe',
    #        'japan', 'asia-pacific-ex-japan', 'north-america', 
    #    ]
    # TODO: regions
    def __init__(self,
                 formed_on: str | list[str] = 'size',
                 sort: str | int | None = None,
    #            region: str = 'us', 
                 **kwargs):
        super().__init__(**kwargs)
        
        #self.region = region

        # formed_on: sorted list of lowercase strings
        if isinstance(formed_on, str):
            self.formed_on = [formed_on.lower()]
        else:
            self.formed_on = sorted([f.lower() for f in formed_on])

        self.sort = str(sort).lower() if sort is not None else self._get_default_sort()
        self.n_portfolios = self._map_sort_to_count()
        self.is_multivariate = len(self.formed_on) > 1

        self._validate_params()


    def _find_table_start(self) -> str:
        """Identify the table section based on weights."""
        return "Equal Weighted Returns" if self.weights == 'ew' else "Value Weighted Returns"


    def _map_sort_to_count(self) -> int:
        """Map sort string/int to total portfolio count."""
        mapping = {
            'tertile': 3, '3': 3,
            'quintile': 5, '5': 5,
            'decile': 10, '10': 10,
            '2x3': 6, '6': 6,
            '5x5': 25, '25': 25,
            '10x10': 100, '100': 100,
            '2x4x4': 32, '32': 32,
        }
        if self.sort not in mapping:
            raise ValueError(f"Sort '{self.sort}' not recognized. Try 'decile', '5x5', etc.")
        return mapping[self.sort]


    def _get_default_sort(self) -> str:
        """Assigns a default sort if none provided."""
        if len(self.formed_on) == 1:
            return 'decile'
        if len(self.formed_on) == 2:
            if any(f in {'ep', 'cfp', 'dp'} for f in self.formed_on):
                return '2x3'
            return '5x5'
        return '2x4x4'


    def _validate_params(self):
        """Validates factors, frequency and sort types."""
        valid = {'size', 'bm', 'op', 'inv', 'ep', 'cfp', 'dp', 'mom',
                 'st_rev', 'lt_rev', 'ac', 'beta', 'ni', 'var', 'resvar'}
        for f in self.formed_on:
            if f not in valid:
                raise ValueError(f"Invalid factor: {f}. Must be one of {sorted(valid)}")

        no_daily = {'ep', 'cfp', 'dp', 'beta', 'ac', 'ni', 'var', 'resvar'}
        if self.frequency == 'd' and any(f in no_daily for f in self.formed_on):
            raise ValueError(f"Daily data is not available for sorts involving {self.formed_on}")

        prior_rets = {'mom', 'st_rev', 'lt_rev'}
        if not self.is_multivariate:
            # Univariate sorts only support 3, 5, 10
            # This fixes eg, "-p 2x3 --on op" returning the full table.
            if self.n_portfolios not in {3, 5, 10}:
                 raise ValueError(
                     f"Sort '{self.sort}' ({self.n_portfolios}) is not a valid univariate sort. "
                     f"Use 'tertile', 'quintile', 'decile' (or 3, 5 or 10)."
                 )

            if self.formed_on[0] in prior_rets: #noqa
                if self.n_portfolios != 10:
                    raise ValueError(f"{self.formed_on[0]} is only available in deciles.")
    
    @override
    def _read(self, data: bytes) -> pa.Table:
        table = super()._read(data)
        
        # fix: drop the "<= 0" column here (bivariates can contain it)
        neg_cols = [name for name in table.column_names if "<=" in name]
        if neg_cols:
            table = table.drop(neg_cols) #

        # Sorts on prior rets are decile only, the rest are a table 
        # of date [+ <= 0] + tertiles, quintiles, deciles.
        if not self.is_multivariate and self.formed_on[0] not in {'mom', 'st_rev', 'lt_rev'}:
            table = self._slice_univariate(table)
        
        return table.combine_chunks()


    def _slice_univariate(self, table: pa.Table) -> pa.Table:
        # No "< =" col now
        mapping = {
            3: slice(1, 4),
            5: slice(4, 9),
            10: slice(9, 19),
        }
        
        target_slice = mapping.get(self.n_portfolios)

        if target_slice is None:
             raise ValueError(f"Cannot slice univariate table: no mapping for {self.n_portfolios} portfolios.")

        if target_slice:
            if table.num_columns > 16:
                indices = [0] + list(range(target_slice.start, target_slice.stop))
                return table.select([i for i in indices if i < table.num_columns])
            # quick fix: 'ac' and 'beta' univariates don't have tertiles... 
            if table.num_columns >= 15:
                adj_mapping = {
                    5: slice(table.num_columns - 15, table.num_columns - 10),
                    10: slice(table.num_columns - 10, table.num_columns),
                }
                adj_slice = adj_mapping.get(self.n_portfolios)
                if adj_slice:
                    indices = [0] + list(range(adj_slice.start, adj_slice.stop))
                    return table.select([i for i in indices if i < table.num_columns])

        return table


    def _get_url(self) -> str:
        """Constructs a URL for a portfolio .zip on the Ken French library."""
        base = "https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp"

        if self.frequency == 'd':
            suffix = "_Daily_CSV.zip"
        elif not self.dividends:
            suffix = "_Wout_Div_CSV.zip"
        else:
            suffix = "_CSV.zip"

        # Univariate sorts
        if not self.is_multivariate:
            mapping = {
                'size': 'Portfolios_Formed_on_ME',
                'bm': 'Portfolios_Formed_on_BE-ME',
                'op': 'Portfolios_Formed_on_OP',
                'inv': 'Portfolios_Formed_on_INV',
                'ep': 'Portfolios_Formed_on_E-P',
                'cfp': 'Portfolios_Formed_on_CF-P',
                'dp': 'Portfolios_Formed_on_D-P',
                'mom': '10_Portfolios_Prior_12_2',
                'st_rev': '10_Portfolios_Prior_1_0',
                'lt_rev': '10_Portfolios_Prior_60_13',
                'ac': 'Portfolios_Formed_on_AC',
                'beta': 'Portfolios_Formed_on_BETA',
                'ni': 'Portfolios_Formed_on_NI',
                'var': 'Portfolios_Formed_on_VAR',
                'resvar': 'Portfolios_Formed_on_RESVAR',
            }
            return f"{base}/{mapping[self.formed_on[0]]}{suffix}"

        # Multivariate sorts
        factors = frozenset(self.formed_on)

        # 3-way sorts  # NOTE REGIONAL AVAIL.
        if len(factors) == 3:
            three_way = {
                frozenset(['size', 'bm', 'op']): "32_Portfolios_ME_BEME_OP_2x4x4",
                frozenset(['size', 'bm', 'inv']): "32_Portfolios_ME_BEME_INV_2x4x4",
                frozenset(['size', 'op', 'inv']): "32_Portfolios_ME_OP_INV_2x4x4",
            }
            slug = three_way.get(factors)
            if not slug: 
                raise ValueError("3-way sort not supported.")

            return f"{base}/{slug}{suffix}"

        # Bivariate sorts
        bivariate_map = {
            frozenset(['size', 'bm']): "",
            frozenset(['size', 'op']): "ME_OP",
            frozenset(['size', 'inv']): "ME_INV",
            frozenset(['bm', 'op']): "BEME_OP",
            frozenset(['bm', 'inv']): "BEME_INV",
            frozenset(['op', 'inv']): "OP_INV",
            frozenset(['size', 'ep']): "ME_EP",
            frozenset(['size', 'cfp']): "ME_CFP",
            frozenset(['size', 'dp']): "ME_DP",
            frozenset(['size', 'ac']): "ME_AC",
            frozenset(['size', 'beta']): "ME_BETA",
            frozenset(['size', 'ni']): "ME_NI",
            frozenset(['size', 'var']): "ME_VAR",
            frozenset(['size', 'resvar']): "ME_RESVAR",
            frozenset(['size', 'mom']): "ME_Prior_12_2",
            frozenset(['size', 'st_rev']): "ME_Prior_1_0",
            frozenset(['size', 'lt_rev']): "ME_Prior_60_13",
        }

        slug_part = bivariate_map.get(factors)
        if slug_part is None:
            raise ValueError(f"Bivariate combination {self.formed_on} not supported.")

        grid_map = {6: ("6_Portfolios", "2x3"),
                    25: ("25_Portfolios", "5x5"),
                    100: ("100_Portfolios", "10x10")}
        prefix, grid = grid_map.get(self.n_portfolios, ("25_Portfolios", "5x5"))

        if not slug_part: # bivariate
            return f"{base}/{prefix}_{grid}{suffix}"
        if "Prior" in slug_part: # bivariate sorts on prior returns
            return f"{base}/{prefix}_{slug_part}{suffix}"

        return f"{base}/{prefix}_{slug_part}_{grid}{suffix}"


class _FamaFrenchIndustryPortfolios(_FFPortfolioBase):
    """Fama-French industry portfolios."""
    _SIZES = {5, 10, 12, 17, 30, 38, 48, 49}

    @property
    def _frequencies(self) -> list[str]: return ['d', 'm', 'y']

    def __init__(self, n_portfolios: int | str = 5, **kwargs):
        n_portfolios = int(n_portfolios) # allow str through
        if n_portfolios not in self._SIZES:
            raise ValueError(f"Industry count must be one of {self._SIZES}")
        self.n_portfolios = n_portfolios
        super().__init__(**kwargs)


    def _find_table_start(self) -> str:
        w_str = "Value Weighted" if self.weights == 'vw' else "Equal Weighted"
        freq_map = {'y': 'Annual', 'd': 'Daily', 'm': 'Monthly'}

        return f"Average {w_str} Returns -- {freq_map.get(self.frequency, 'Monthly')}"


    def _get_url(self) -> str:
        base = "https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp"

        if self.frequency == 'd':
            suffix = "daily_CSV.zip"
        else:
            suffix = "Wout_Div_CSV.zip" if not self.dividends else "CSV.zip"

        return f"{base}/{self.n_portfolios}_Industry_Portfolios_{suffix}"


def _get_ff_portfolios(
    formed_on: str | list[str] = 'size', #on=, sorted_on=, might rename
    sort: str | int | None = None, #2x3, decile, 25
    industry: int | None = None, #industries= ? allow str through, "12" 
    weights: str = 'vw',
    frequency: str = 'm',
    **kwargs) -> pa.Table:
    """Internal helper: factory to return.

    Factory for Fama French portfolios. Either `industry` or 
    `formed_on` required.
    """
    # we industry?
    if industry is not None:
        return _FamaFrenchIndustryPortfolios(
            frequency=frequency, weights=weights, n_portfolios=industry, **kwargs
        )

    # keep formed_on None, let class handle it.
    return _FamaFrenchSorts(
        frequency=frequency, weights=weights, formed_on=formed_on, sort=sort, **kwargs
    )

# TODO: setters/getters, regions, tests, aggregates for q? prob not. 



#most models do this:
(main, aqr, icr liq, misp, q)
except (pa.ArrowIOError, pa.ArrowInvalid) as e:
            msg = f"{self.__class__.__name__}: reading failed: {e}"
            self.log.error(msg)
            raise ValueError(msg) from e


(pa.ArrowIOError, pa.ArrowInvalid)
ArrowInvalid
ValueError
KeyError
(ValueError, TypError)
ImportError (pl, pd)
PackageNotFoundError (cli, version)

httpx.HTTPStatusError
httpx.RequestError

few catchall Exception



# https://github.com/encode/httpx/blob/master/httpx/_exceptions.py



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import io
from typing import Any
import pyarrow as pa
from python_calamine import CalamineWorkbook
from getfactormodels.models.base import FactorModel
from getfactormodels.models.fama_french import FamaFrenchFactors
from getfactormodels.utils.arrow_utils import (
    round_to_precision,
    scale_to_decimal,
    select_table_columns,
)
from getfactormodels.utils.date_utils import offset_period_eom


class DHSFactors(FactorModel):
    """The Behavioural Factors of Daniel, Hirshleifer and Sun (2020).
    
    FIN and PEAD factors. Data from July 1972 - December, 2023.

    References:
        K. Daniel, D. Hirshleifer and L. Sun, 2020. Short and Long Horizon
        Behavioral Factors. Review of Financial Studies, 33 (4): 1673-1736.

    """
    # copyright/attribution info! TODO
    @property
    def _frequencies(self) -> list[str]: return ['d', 'm']
    
    @property
    def _precision(self) -> int: return 12

    @property  
    def schema(self) -> pa.Schema:
        date = [('Date', pa.string())]   # 197202
        fin = [('FIN', pa.float64())]
        pead = [('PEAD', pa.float64())]
        
        # FIN and PEAD, swap cols between frequencies.
        if self.frequency == 'd':
            return pa.schema(date + fin + pead)
        return pa.schema(date + pead + fin)


    def _get_url(self) -> str:
        base_url = 'https://docs.google.com/spreadsheets/d/'

        if self.frequency == 'd':
            sheet_id = '1lWaNCuHeOE-nYlB7GA1Z2-QQa3Gt8UJC'
            #info_id =
        else:
            sheet_id = '1VwQcowFb5c0x3-0sQVf1RfIcUpetHK46'
            #info_sheet_id = '#gid=96292754'  # Construction, Universe, Period, More Details

        return  f'{base_url}{sheet_id}/export?format=xlsx'


    # Precision fix: export as .xlsx, then read with Calamine. 
    def _read(self, data: bytes) -> pa.Table:
        wb = CalamineWorkbook.from_filelike(io.BytesIO(data))

        rows = wb.get_sheet_by_name(wb.sheet_names[0]).to_python()
        headers = [str(h).strip() for h in rows[0]]

        _dict = {name: [row[i] for row in rows[1:]] for i, name in enumerate(headers)}

        # fix: load without schema here
        table = pa.Table.from_pydict(_dict)

        # Enforce schema here
        table = table.select(self.schema.names).cast(self.schema)
        
        table = scale_to_decimal(table)
        table = offset_period_eom(table, self.frequency)
        
        if "Date" in table.column_names:
            table = table.rename_columns(["date"] + table.column_names[1:])

        # Wrap in FF Mkt-RF and RF
        _ff = FamaFrenchFactors(model='3', 
                                frequency=self.frequency).load()
        _ff = select_table_columns(_ff.data, ['Mkt-RF', 'RF'])

        return table.join(_ff, keys="date", join_type="inner").combine_chunks()



# getfactormodels/registry.py
from types import MappingProxyType
from getfactormodels.models.base import FactorModel
from getfactormodels.models.fama_french import FamaFrenchFactors, FFIndustryPortfolios
from getfactormodels.models.carhart import CarhartFactors
from getfactormodels.models.q_factors import QFactors
from getfactormodels.models import *

# Maps canonical keys to lists of user-friendly variations
_MODEL_ALIASES = {
    # Factors
    "3": ["3", "ff3", "famafrench3"],
    "4": ["4", "ff4", "carhart", "car"],
    "5": ["5", "ff5", "famafrench5"],
    "6": ["6", "ff6", "famafrench6"],
    "Q": ["q", "qfactors", "q-factors", "q_factors", "q5", "hmxz"],
    "Qclassic": ["q4", "qclassic", "q-classic", "q_classic", "classic_q"],
    "HMLDevil": ["hmld", "hmldevil", "hml_devil", "devil"],
    "QMJ": ["qmj", "quality", "qualityminusjunk"],
    "BAB": ["bab", "betting", "bettingainstbeta"],
    "Mispricing": ["mispricing", "mis", "misp"],
    "Liquidity": ["liq", "liquidity"],
    "ICR": ["icr", "intermediary", "hkm"],
    "DHS": ["dhs", "behavioural", "behaviour"],
    "BarillasShanken": ["bs", "bs6", "barillasshanken", "barillas-shanken"],
    "VME": ["vme", "valmom", "valueandmomentumeverywhere"],
    "AQR6": ["aqr6", "aqr", "aqrfactors"],
    "HighIncomeCCAPM": ["hccapm", 'hcapm', 'hc-capm'],
    "ConditionalCAPM": ["jwcapm", "plcapm", "jwccapm", "plccapm", "ccapm"],
    # FF Portfolios
    "Industry": ["industry", "ind", "sector"],
}

# Invert aliases for O(1) lookup: 'ff3' -> '3'
_ALIAS_LOOKUP = {
    alias: key 
    for key, aliases in _MODEL_ALIASES.items() 
    for alias in aliases
}

# map key to the class
_CLASS_REGISTRY = {
    "3": FamaFrenchFactors,
    "4": CarhartFactors,
    "5": FamaFrenchFactors,
    "6": FamaFrenchFactors,
    "Q": QFactors,
    "Qclassic": QFactors,
    "Industry": FFIndustryPortfolios,
}

def resolve_model_key(user_input: str | int) -> str:
    """
    Normalize user input to a canonical key.
    Example: 'ff3' -> '3', 'industry' -> 'Industry'
    """
    if not user_input:
        raise ValueError("Model name cannot be empty.")
        
    val = str(user_input).lower().strip().replace("-", "").replace("_", "")
    
    # Direct match or alias match
    if val in _ALIAS_LOOKUP:
        return _ALIAS_LOOKUP[val]
        
    # Fallback: if the user passes "3" (int) or strict key
    for key in _CLASS_REGISTRY:
        if val == key.lower():
            return key
            
    raise ValueError(f"Unknown model or portfolio: '{user_input}'")

def get_model_class(key: str) -> "FactorModel":
    """Retrieve the class constructor for a resolved key."""
    if key not in _CLASS_REGISTRY:
        raise ValueError(f"Model key '{key}' is valid but not registered to a class.")
    return _CLASS_REGISTRY[key]
